/* $Id: tspAppInit.cc,v 1.2 1996/07/04 08:13:23 uwe Exp $ */
/* 
 * tkAppInit.c --
 *
 *	Provides a default version of the Tcl_AppInit procedure for
 *	use in wish and similar Tk-based applications.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tkAppInit.c 1.20 96/02/15 18:55:27
 */

#include "tcl.h"
#include "tk.h"

/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */

// extern int matherr();
// int *tclDummyMathPtr = (int *) matherr;
 
// #ifdef TK_TEST
// EXTERN int		Tktest_Init _ANSI_ARGS_((Tcl_Interp *interp));
// #endif /* TK_TEST */

/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *	This is the main program for the application.
 *
 * Results:
 *	None: Tk_Main never returns here, so this procedure never
 *	returns either.
 *
 * Side effects:
 *	Whatever the application does.
 *
 *----------------------------------------------------------------------
 */

int main(int argc, char **argv)
    	/* Number of command-line arguments. */
  		/* Values of command-line arguments. */
{
    Tk_Main(argc, argv, Tcl_AppInit);

    return 0;			/* Needed only to prevent compiler warning. */
}


/***************************************************************************/
/***************************************************************************/
/* File: gentsp.cpp (Eigenproduktion)                                      */

#include "tsp.hpp"
#include <time.h>
#include <assert.h>
#include <stdlib.h>

#define BOOL int


/***************************************************************************/
/* Parameter fuer Genetische Algorithmen                                   */

int PopulGroesse = 200;  // Populationsgroesse
int ReproRate    = 18;   // Reproduktionsrate (9%)
int MutatRate    = 2;    // Mutationsrate (1%)
int CORate       = 120;  // CrossOverRate (60%)
int SelRate      = 60;   // Selektionsrate (30%) = Die ... Besten


/***************************************************************************/
/* Ein Chromosom mit zufaelligen Werten von 0.00 - 0.99 initialisieren.    */
/* Parameter : Laenge des Chromosoms, das Chromosom (Ref.Par.)             */
/* Funktionswert : keiner                                                  */

void InitChromosom(int ChrLaenge, CHROMOSOM & Chromo) {

  for (int i=1; i<=ChrLaenge; i++) {
	 REAL ZufZahl;
	 // Zufallszahl erzeugen zwischen 0.00 und 0.99
	 ZufZahl=(float)rand()/(float)RAND_MAX;
	 // Zahl an das Chromosom anhaengen
	 Chromo.appr(ZufZahl);
  }
}


/***************************************************************************/
/* Anfangspopulation fuer die genetischen Algorithmen initialisieren.      */
/* Parameter : Laenge der Chromosome, Population (Ref.Par.)                */
/* Funktionswert : keiner                                                  */

void InitPopulation(int ChrLaenge, POPULATION & Popul) {

  for (int i=1; i<=PopulGroesse; i++) {
	 CHROMOSOM Chromo(empty);
	 // Ein Chromosom herstellen
	 InitChromosom(ChrLaenge, Chromo);
	 // Chromosom an die Population anhaengen
	 Popul.appr(Chromo);
  }
}


/***************************************************************************/
/* Aus dem Chromosom den Weg berechnen = Sortierung der Gene ermitteln.    */
/* Parameter : Das Chromosom, Laenge des Chromosom, Loesungsweg (Point.)   */
/* Funktionswert : keiner                                                  */

void CalcPath(const CHROMOSOM & Chromo, int ChrLaenge, int* Loesung) {

  BOOL* Taken;                 // Feld: Gen ausgewertet(J/N) ?
  REAL Kleinster;              // Kleinster Wert im Chromosom
  unsigned int KleinstePos;    // Pos. des kleinsten Wertes im Chromosom
  int i;		                   // Laufvariable

  // Speicher fuer BOOLVektor alloziieren und Vektor initialisieren
  Taken=new BOOL[ChrLaenge];
  assert(Taken!=0);
  for (i=0; i<ChrLaenge; i++) Taken[i]=FALSE;
  for (i=0; i<ChrLaenge; i++) {
	 // Kleinster mit Wert groesser als alle Gene (max. 1.0) initialisieren
	 Kleinster=1.5;
	 // Chromosom durchlaufen und kleinstes Gen suchen
	 for (int j=1; j<=ChrLaenge; j++) {
		if (Chromo.sel(j)<Kleinster && Taken[j-1]==FALSE) {
		  Kleinster=Chromo.sel(j);
		  KleinstePos=j;
		}
	 }
	 // Reiseziel festlegen (+1, da Punkt 1 = Start = nicht variabel)
	 Loesung[i]=KleinstePos+1;
	 // Bereits gewaehltes Gen blockieren
	 Taken[KleinstePos-1]=TRUE;
  }
  delete [] Taken;
}


/***************************************************************************/
/* Entfernungsmatrix mit den Entfernungswerten initialisieren.             */
/* Parameter : String mit allen Werten (P), Entfernungsmatrix (RP)         */
/* Funktionswert : Dimension der Matrix = Startpunkt der Reise             */

int InitTable(char* Uebergabe, ZEILE & Tabelle) {

  int i=0;
  int Anzahl;
  REAL entfernung;
  float input;

  // Datum Nr.1 lesen = Dimension der Entfernungsmatrix
  while (Uebergabe[i]!=' ') i++;
  sscanf(Uebergabe,"%d",&Anzahl);
  // Daten Nr.2 bis Nr.(Anzahl^2+1) lesen = Daten der Entfernungsmatrix
  for (int j=1; j<=Anzahl; j++) {
	 SPALTE spalte(empty);
	 for (int k=1; k<=Anzahl; k++) {
		// Zeiger zum naechsten Datum setzen
		Uebergabe=Uebergabe+(i+1);
		i=0;
		// Nur weiterlaufen, wenn nicht letztes Element
		if ( ! (j==Anzahl && k==Anzahl))
		  while (Uebergabe[i]!=' ') i++;
		// Datum lesen und in die Matrix eintragen
		sscanf(Uebergabe,"%f",&input);
		entfernung=input;
		spalte.appr(entfernung);
	 }
	 Tabelle.appr(spalte);
  }
  return Anzahl;
}


/***************************************************************************/
/* Laenge des ausgewaehlten Pfades anhand der Entfernungsmatrix berechnen. */
/* Parameter : Laenge des Chromosoms, Loesungsvektor (P),                  */
/*             Entfernungsmatrix, Startpunkt der Reise                     */
/* Funktionswert : Laenge der Reise                                        */

float CalcLength(int ChrLaenge, int* Loesung, const ZEILE & Tabelle) {

  float result=0;           // Ergebnis der Weglaengenberechnungen
  SPALTE spalte(empty);     // Hilfsvariable
  REAL erg;                 // Hilfsvariable
  int i;                    // Laufvariable

  // Vom Start zum ersten Punkt
  spalte=Tabelle.sel(1);
  erg=spalte.sel(Loesung[0]);
  result+=erg.cv_real();
  // Vom ersten bis zum letzten Punkt
  for (i=1; i<ChrLaenge; i++) {
	 spalte=Tabelle.sel(Loesung[i-1]);
	 erg=spalte.sel(Loesung[i]);
	 result+=erg.cv_real();
  }
  // Vom letzten Punkt zum Start zurueck
  spalte=Tabelle.sel(Loesung[i-1]);
  erg=spalte.sel(1);
  result+=erg.cv_real();
  return result;
}


/***************************************************************************/
/* Das beste Chromosom bzgl. Weglaenge aus der Population suchen.          */
/* Parameter : Population, Chromosomenlaenge, Entfernungsmatrix,           */
/*             Das beste Chromosom                                         */
/* Funktionswert : Index des besten Chromosom                              */

CHROMOSOM GetBest(const POPULATION & Popul, int ChrLaenge,
						const ZEILE & Tabelle) {

  CHROMOSOM BestChr(empty); // Bestes Chromosom
  float BestWay,ZwErg;      // Kuerzester Reiseweg + Zwischenergebnis
  int* Loesung;             // Loesungsvektor

  // Speicher fuer Loesungsvektor alloziieren
  Loesung=new int[ChrLaenge];
  assert(Loesung!=0);
  // Erstes Chromosom als Bestes festlegen
  BestChr=Popul.sel(1);
  CalcPath(BestChr,ChrLaenge,Loesung);
  // Weg berechnen
  BestWay=CalcLength(ChrLaenge,Loesung,Tabelle);
  // Die weiteren Chromosome der Population durchlaufen,
  // ueberprufen, ob besser und bestes Chromosom evtl. ersetzen
  for (int i=2; i<=Popul.len().cv_nat0(); i++) {
	 CalcPath(Popul.sel(i),ChrLaenge,Loesung);
	 ZwErg=CalcLength(ChrLaenge,Loesung,Tabelle);
	 if (ZwErg<BestWay) {
		BestWay=ZwErg;
		BestChr=Popul.sel(i);
	 }
  }
  delete [] Loesung;
  return BestChr;
}


/***************************************************************************/
/* Loesungvektor fuer Tcl/Tk in String konvertieren                        */
/* Parameter : Loesungsstring (P), Loesungsvektor (P),                     */
/*             Chromosomenlaenge, Weglaenge                                */
/* Funktionswert : keiner                                                  */

void LoesungToString(char* String, int* Loesung, int ChrLaenge, float Laenge){

  char* pAnfang;     // Merkzeiger

  // Weglaenge in den String schreiben
  sprintf(String,"%f",Laenge);
  strcat(String," ");
  pAnfang=String;
  // Die anzufahrenden Punkte in richtiger Reihenfolge in den String schreiben
  for (int i=0; i<ChrLaenge; i++) {
	 String+=strlen(String);
	 sprintf(String,"%d",Loesung[i]);
	 strcat(String," ");
  }
  // Zeiger zurueck auf den Stringanfang setzen
  String=pAnfang;
}


/***************************************************************************/
/* Zufaellige Auswahl eines Chromosoms fuer die Vererbungsmechanismen.     */
/* Prinzip: - Zwei Chromsomen zufaellig auswaehlen                         */
/*          - Chromosom mit besserer Fitness wird zurueckgeliefert         */
/* Parameter : Population, Chromosomenlaenge, Entf.Matrix                  */
/* Funktionswert : Das Chromosom                                           */

CHROMOSOM SelectChromo(const POPULATION & Popul, int ChrLaenge,
							  const ZEILE & Tabelle) {

  unsigned int Zuffi_1,Zuffi_2;    // Zufallszahlen
  POPULATION HilfsPop(empty);      // Hilfspopulation

  // Zwei Zufallszahlen erzeugen und sicherstellen, dass nicht gleich
  do {
	 Zuffi_1=(rand()%PopulGroesse)+1;
	 Zuffi_2=(rand()%PopulGroesse)+1;
  }
  while (Zuffi_1==Zuffi_2);
  // Hilfspopulation mit den 2 gewaehlten Chromosomen erzeugen
  HilfsPop.appr(Popul.sel(Zuffi_1));
  HilfsPop.appr(Popul.sel(Zuffi_2));
  // Chromosom mit besserer Fitness auswaehlen
  return GetBest(HilfsPop,ChrLaenge,Tabelle);
}


/***************************************************************************/
/* Ein Verebungsschritt durchfuehren ==> Neue Population erzeugen.         */
/* Parameter : Population, Chromosomenlaenge, Entfernungsmatrix            */
/* Funktionswert : Die neue Population                                     */

POPULATION Vererbung(POPULATION Popul, int ChrLaenge, const ZEILE & Tabelle) {

  POPULATION PopulNeu(empty);                     // Neue Population
																  // Chromosomen fuer Auswahl
  CHROMOSOM Chromo(empty), Chromo_1(empty), Chromo_2(empty);
  unsigned int iZuffi;                            // Zufallszahlen
  REAL rZuffi;                                    //      -"-
  int i;                                          // Laufvariable
  FITNESS Fit(empty);	                          // Fitnessfeld
  REAL FitWert;                                   // Fitnesswert
  int *Loesung;                                   // Loesungsvektor
  unsigned int Index;                             // Hilfsindex

  // Fitness fuer alle Chromosomen berechnen und speichern
  Loesung=new int[ChrLaenge];
  assert(Loesung!=0);
  for (i=1; i<=PopulGroesse; i++) {
	 CalcPath(Popul.sel(i),ChrLaenge,Loesung);
	 FitWert=CalcLength(ChrLaenge,Loesung,Tabelle);
	 Fit.appr(FitWert);
  }
  delete [] Loesung;
  // Reproduktion
  for (i=1; i<=ReproRate; i++) {
	 // Chromosom einfach anhaengen
	 PopulNeu.appr(SelectChromo(Popul,ChrLaenge,Tabelle));
  }
  // Mutation
  for (i=1; i<=MutatRate; i++) {
	 // Chromosom auswaehlen
	 Chromo=SelectChromo(Popul,ChrLaenge,Tabelle);
	 // Gen fuer die Mutation per Zufall bestimmen
	 iZuffi=(rand()%ChrLaenge)+1;
	 // Neuen Wert fuer Gen per Zufall bestimmen
	 rZuffi=(float)rand()/(float)RAND_MAX;
	 Chromo.ovwrt1(iZuffi,rZuffi);
	 PopulNeu.appr(Chromo);
  }
  // Crossover
  for (i=1; i<=CORate; i++) {
	 // Chromosom 1 bestimmen
	 Chromo_1=SelectChromo(Popul,ChrLaenge,Tabelle);
	 // Chromosom 2 bestimmen
	 Chromo_2=SelectChromo(Popul,ChrLaenge,Tabelle);
	 // (Keine Ueberpruefung, ob zweimal das gleiche Chromosom gewaehlt wird,
	 //  da unwahrscheinlich und erheblicher Overhead.)
	 // Startgen fuer Crossover bestimmen
	 iZuffi=(rand()%ChrLaenge)+1;
	 // Genteile zusammenfuegen
	 Chromo=Chromo_1.sub_tup(1,iZuffi)
			 +Chromo_2.sub_tup(iZuffi+1,ChrLaenge);
	 PopulNeu.appr(Chromo);
  }
  // Selektion
  for (i=1; i<=SelRate; i++) {
	 Index=1;
	 FitWert=Fit.sel(1);
	 // Bestes Chromosom heraussuchen
	 for (int j=2; j<=Popul.len().cv_nat0(); j++) {
		if (Fit.sel(j)<FitWert) {
		  Index=j;
		  FitWert=Fit.sel(j);
		}
	 }
	 // Chromosom anhaengen und aus Population loeschen
	 PopulNeu.appr(Popul.sel(Index));
	 Popul.rmv1(Index);
	 Fit.rmv1(Index);
  }
  return PopulNeu;
}
/***************************************************************************/
/* Das beste Chromosom als String zurueckliefern.                          */
/* Parameter : Population, Chromosomenlaenge, Entfernungsmatrix            */
/* Funktionswert : Die neue Population                                     */

void GetBestString(const POPULATION & Popul, const ZEILE & Tabelle,
		   int ChrLaenge, char* LoesStr) {

  CHROMOSOM Chromo(empty);          // Ein Chromosom
  float WegLaenge;                  // Laenge des Reiseweges
  int* Loesung;                     // Loesungsvektor

  Loesung=new int[ChrLaenge];
  assert(Loesung!=0);
  // Bestes Chromosom bestimmen
  Chromo=GetBest(Popul,ChrLaenge,Tabelle);
  // Weg des besten Chromosoms errechnen
  CalcPath(Chromo,ChrLaenge,Loesung);
  // Reiselaenge bestimmen
  WegLaenge=CalcLength(ChrLaenge,Loesung,Tabelle);
  // Route in String konvertieren
  LoesungToString(LoesStr,Loesung,ChrLaenge,WegLaenge);
  delete [] Loesung;
}


/***************************************************************************/
/***************************************************************************/
/* SCHNITTSTELLE ZU TCL/TK                                                 */

int Connect(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {

  static POPULATION Popul;          // Population
  static ZEILE Tabelle;             // Entfernungsmatrix
  static int ChrLaenge=0;           // Laenge der Chromosomen
  static int Anzahl=0;              // Dimension des Entfernungsmatrix

  char* LoesStr;                    // Loesungsvektor in String-Form
  time_t t;                         // Zeit fuer Zuf.Zahlen.Gen.Init.
  int Aktion;                       // Aktionscode
  char* zeiger;                     // Laufzeiger
  int i;                            // Laufvariable

  // Minimum der Argumente oK?
  if (argc<2) {
	 interp->result="wrong # args";
	 return TCL_ERROR;
  }
  // Genetische Parameter korrekt ?
  assert(ReproRate+MutatRate+CORate+SelRate==PopulGroesse);
  // 1. Argument lesen = Aktionscode
  sscanf(argv[1],"%d",&Aktion);
  // Verzweigung in Abhaengigkeit vom Aktionscode
  switch(Aktion) {
	 case 1: // INITIALISIERUNG
			// Zufallszahlengenerator initialisieren
			srand((unsigned)time(&t));
			// Entfernungsmatrix initialisieren
                        Tabelle=empty;
			Anzahl=InitTable(argv[2],Tabelle);
			// Init. Chromosomenlaenge (kleiner wegen Startpunkt)
			ChrLaenge=Anzahl-1;
			// Population initialisieren
			Popul=empty;
                        InitPopulation(ChrLaenge,Popul);
			LoesStr=new char[ChrLaenge*4];
			assert(LoesStr!=0);
			// Beste Route als String erhalten
			GetBestString(Popul,Tabelle,ChrLaenge,LoesStr);
			// Ergebnis an Tcl/Tk
			strcpy(interp->result,LoesStr);
			delete [] LoesStr;
			return TCL_OK;
	 case 2: // GENETISCHE PARAMETER SETZEN
				zeiger=argv[2];
				// Populationsgroesse einlesen
				sscanf(zeiger,"%d",&PopulGroesse);
				while (*zeiger!=' ') zeiger++;
                                zeiger++;
				// Reproduktionsrate einlesen
				sscanf(zeiger,"%d",&ReproRate);
				while (*zeiger!=' ') zeiger++;
                                zeiger++;
				// Mutationsrate einlesen
				sscanf(zeiger,"%d",&MutatRate);
				while (*zeiger!=' ') zeiger++;
                                zeiger++;
				// CrossOverrate einlesen
				sscanf(zeiger,"%d",&CORate);
				while (*zeiger!=' ') zeiger++;                                
				// Selektionsrate einlesen
				sscanf(zeiger,"%d",&SelRate);
				return TCL_OK;
	 case 3: // VEREBUNG ANTOSSEN
				// Vererbungsschritt durchfuehren
				Popul=Vererbung(Popul,ChrLaenge,Tabelle);
				LoesStr=new char[ChrLaenge*4];
				assert(LoesStr!=0);
				// Beste Route als String erhalten
				GetBestString(Popul,Tabelle,ChrLaenge,LoesStr);
				// Ergebnis an Tcl/Tk
				strcpy(interp->result,LoesStr);
				delete [] LoesStr;
				return TCL_OK;
  }
}

/***************************************************************************/
/***************************************************************************/

/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int Tcl_AppInit(Tcl_Interp *interp)
    	/* Interpreter for application. */
{
    if (Tcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);

/*
#ifdef TK_TEST
    if (Tktest_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
#endif 
*/

    Tcl_CreateCommand(interp, "tspconnect", Connect, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

//  tcl_rcFileName="~/wishrc";


    Tcl_SetVar(interp, "tcl_rcFileName", "~/.wishrc", TCL_GLOBAL_ONLY);

    return TCL_OK;
}
